#!/usr/bin/env ruby
# potrebuje pcap_read - pcap 0.9.8, 1.0 uz nefunguje
# pcap-ruby nema sendpacket

# TODO allow sendpacket

require 'gtk2'
require 'pcap'
require 'macaddr'

# aging timeout (seconds)
AGING = 300
# check aging (milliseconds)
AGING_CHECK = 1000

IGNORED = nil
IN_BLOCK = "red"
OUT_BLOCK = "orange"
PASSED = "green"
FILTERED = "yellow"

# index of color in iter
COLOR = 8

# convert n-byte number to byte string
def num2str(num, bytes)
  a = []
  bytes.times do |i|
    a.push(num & 255)
    num = num >> 8
  end
  #a.pack("C#{bytes}")
  a.reverse.pack("C*")
end


class Pcap::Packet
  def mac_src
    raw_data[6,6]
  end

  def mac_dst
    raw_data[0,6]
  end

  def broadcast?
    return mac_dst == "\377\377\377\377\377\377" #num2str(0xffffffffffff, 6)
  end

  def icmp?
    return (ip? and ip_proto == 1)
  end

  # dump packet (returns string)
  def dump
    if ip?
      if tcp?
        buf = "TCP\n"
      elsif udp?
        buf = "UDP\n"
      elsif ip_proto == 1
        buf = "ICMP\n"
      else
        buf = "IP\n"
      end
    else
      buf = ''
    end
    #if self.tcp?
    #  #buf << " (%s)" % self.tcp_data_len
    #  buf << " ack %d" % self.tcp_ack if self.tcp_ack?
    #  buf << " win %d" % self.tcp_win
    #end
    #if self.ip?
    #  buf << " (DF)" if self.ip_df?
    #end
    #buf << "\n"
    #

    cnt = 0
    while true
      buf << sprintf("%04d: ", cnt)
      line = raw_data[cnt, 16]
      line.each_byte { |c|
        buf << sprintf("%02x ", c)
      }
      buf << sprintf("\n")

      break if line.length < 16
      cnt += 16
    end
    #puts buf
    return buf
  end
end


class LookupEntry < Array
  def initialize(p, s)
    super(2)
    self[0] = p
    self[1] = s
  end

  def port=(val)
    self[0] = val
  end

  def port
    self[0]
  end

  def stamp=(val)
    self[1] = val
  end

  def stamp
    self[1]
  end
end


# mac => port lookup table with entry aging
class LookupTable < Hash
  def initialize(win)
    super()
    @window = win
    @window.stats.each { |e| e.text = '0' }
    @l2_cnt = @icmp_cnt = @ip_cnt = @udp_cnt = @tcp_cnt = 0

    # mutex for table
    @mutex = Mutex.new

    # start aging check
    @continue = true
    GLib::Timeout.add(AGING_CHECK) {
      if @continue
        @mutex.synchronize {
          self.each_pair { |k, v|
            if v.stamp+AGING <= Time.now.to_i
              @window.macstore.each { |model, path, iter|
                if sprintf("%02X:%02X:%02X:%02X:%02X:%02X", *k.unpack("C6")) ==
                    iter[1]
                  @window.macstore.remove(iter)
                  break
                end
              }

              self.delete(k)
            end
          }
        }
      end
      @continue
    }
  end

  # stop aging check
  def stop
    @continue = false
  end

  def update_counters(pkt, ignored)
    if ignored
      $ignored += 1
      @window.stats[5].text = $ignored.to_s
    end

    @l2_cnt += 1
    @window.stats[0].text = @l2_cnt.to_s

    if pkt.icmp?
      @icmp_cnt += 1
      @window.stats[1].text = @icmp_cnt.to_s
    elsif pkt.ip?
      @ip_cnt += 1
      @window.stats[2].text = @ip_cnt.to_s
    end

    if pkt.udp?
      @udp_cnt += 1
      @window.stats[3].text = @udp_cnt.to_s
    end

    if pkt.tcp?
      @tcp_cnt += 1
      @window.stats[4].text = @tcp_cnt.to_s
    end
  end

  # learn MAC, update GUI and save packet
  def learn(pkt, port)
    # elsewhere since we added ignore counter
    # update_counters(pkt)

    src = pkt.mac_src
    @mutex.synchronize {
      n = self[src]
      if n                      # existing entry
        n.port = port
        n.stamp = Time.now.to_i
        @window.macstore.each { |model, path, iter|
          if sprintf("%02X:%02X:%02X:%02X:%02X:%02X", *src.unpack("C6")) ==
              iter[1]
            iter[2] = n.port.dev
            iter[3] = n.stamp
            break
          end
        }
      else                      # new entry
        entry = LookupEntry.new(port, Time.now.to_i)
        self[src] = entry

        iter = @window.macstore.append
        iter[0] = entry
        iter[1] = sprintf("%02X:%02X:%02X:%02X:%02X:%02X", *src.unpack("C6"))
        iter[2] = port.dev
        iter[3] = entry.stamp
      end
    }

    # save packet
    iter = @window.packetstore.append
    iter[0] = pkt
    iter[1] = sprintf("%02X:%02X:%02X:%02X:%02X:%02X", *pkt.mac_dst.unpack("C6"))
    iter[2] = sprintf("%02X:%02X:%02X:%02X:%02X:%02X", *pkt.mac_src.unpack("C6"))
    iter[3] = sprintf("%02X%02X", *pkt.raw_data[12,2].unpack("C2"))
    if pkt.ip?
      iter[4] = pkt.ip_dst.to_s
      iter[6] = pkt.ip_src.to_s

      if pkt.tcp? or pkt.udp?
        iter[5] = pkt.dport
        iter[7] = pkt.sport
      end
    end

   return iter
  end

  # dump lookup table
  def dump
    each_pair { |k, v|
      #k.each_byte { |b| printf("%02x", b) }
      printf("%02x:%02x:%02x:%02x:%02x:%02x", *k.unpack("C6"))
      puts " => #{v.port.dev} (#{v.stamp})"
    }
  end
end


class Filter
  attr_reader :policy_in, :policy_out

  def initialize
    # TODO: nie, nebudeme pouzivat local IP ignore, len MAC
    @ignore_ip = {}
    @ignore_mac = $local_mac
    clear
  end

  def clear
    @in = []
    @out = []
    # default policy: allow, match is denied
    @policy_in = @policy_out = true
  end

  def add(rule)
    if rule.direction == 'in'
      if rule.policy == nil
        @in.push(rule)
      else
        @policy_in = rule.policy
      end
    else
      if rule.policy == nil
        @out.push(rule)
      else
        @policy_out = rule.policy
      end
    end
  end

  def match(list, pkt, policy)
    list.each { |r|
      return policy if r.match(pkt)
    }
    return !policy
  end

  # let packet out?
  def out?(pkt)
    return !match(@out, pkt, @policy_out)
  end

  # let the packet in?
  def in?(pkt)
    return !match(@in, pkt, @policy_in)
  end

  def ignore?(pkt)
    return (match_mac(pkt.mac_src, pkt.mac_dst))
    # or (pkt.ip? and match_ip(pkt.ip_src, pkt.ip_dst)))
  end

  def match_mac(src, dst)
    @ignore_mac.each { |v|
      return true if v == src or v == dst
    }
    return false
  end

  def match_ip(src, dst)
    @ignore_ip.each_value { |v|
      return true if v[0] == src or v[0] = dst
    }
    return false
  end
end


class Port
  attr_reader :pcap, :dev, :filter, :store

  def initialize(dev, index, portstats)
    @dev = dev
    @filter = Filter.new
    # ignore only MAC
    # @filter.add(dev, Pcap.lookupnet(dev))

    # starting index for portstats
    @index = index
    @stats = portstats
    ['in', 'out'].each { |dir|
      iter = @stats.append
      iter[0] = [dir, dev].join(' ')
      1.upto(6) { |i| iter[i] = 0 }
    }

    @pcap = Pcap::Capture.open_live(dev, 2000, true)
  end

  def start
    @thread = Thread.new {
      loop
    }
  end

  def stop
    @thread.kill
  end

  def update_stats(stats, pkt)
    stats[1] += pkt.size
    stats[2] += 1
    if pkt.icmp?
      stats[3] += 1
    elsif pkt.ip?
      stats[4] += 1
      stats[5] += 1 if pkt.tcp?
      stats[6] += 1 if pkt.udp?
    end
  end

  def loop
    @pcap.each_packet { |pkt|
      iter = $table.learn(pkt, self)

      if @filter.ignore?(pkt)
        $table.update_counters(pkt, true)
        iter[COLOR] = IGNORED
      else
        $table.update_counters(pkt, false)
        iter[COLOR] = IN_BLOCK
        # if allowed to accept by rules
        if @filter.in?(pkt)
          # keep input stats
          update_stats(@stats.get_iter(@index.to_s), pkt)
          # forward to appropriate ports
          result = forward(pkt)

          #p result
          if result.class == Array
            if result.include?(false)
              iter[COLOR] = OUT_BLOCK
            elsif result.include?(true)
              iter[COLOR] = PASSED
            elsif result.include?(nil)
              iter[COLOR] = FILTERED
            end
          else # forwarded
            case result
            when nil then iter[COLOR] = FILTERED
            when true then iter[COLOR] = PASSED
            when false then iter[COLOR] = OUT_BLOCK
            end
          end
        end
      end
    }
  end

  def forward(pkt)
    # broadcast or unknown destination (FLOODING)
    if pkt.broadcast? || !$table[pkt.mac_dst]
      # all but source port
      return $ports.collect { |p|
        if p != self
          p.send(pkt)
        end
      }
    else
      port = $table[pkt.mac_dst].port
      # only if src/dst are not on the same segment (FILTERING and FORWARDING)
      if self != port
        return $table[pkt.mac_dst].port.send(pkt)
      else
        return nil
      end
    end
  end

  def send(pkt)
    # if allowed to send by rules
    if @filter.out?(pkt)
      # keep output stats
      update_stats(@stats.get_iter((@index+1).to_s), pkt)
# TODO allow: uncomment this line:
       @pcap.sendpacket(pkt.raw_data, pkt.raw_data.length)
      return true
    else
      return false
    end
  end

  def close
    @pcap.close
  end
end


class MainWindow
  attr_reader :macstore, :packetstore, :stats

  def initialize
    reset_tables

    @window = Gtk::Window.new(Gtk::Window::TOPLEVEL)
    @window.resizable = true
    @window.title = "PS2 Switch"
    @window.border_width = 10
    @window.signal_connect('delete_event') { Gtk.main_quit }
    @window.set_size_request(800, 500)#-1)

    # pridanie treeview
    @macview = Gtk::TreeView.new(@macstore)
    renderer = Gtk::CellRendererText.new
    # Add column using our renderer
    col = Gtk::TreeViewColumn.new("MAC", renderer, :text => 1)
    @macview.append_column(col)
    col = Gtk::TreeViewColumn.new("Port", renderer, :text => 2)
    @macview.append_column(col)
    col = Gtk::TreeViewColumn.new("Timestamp", renderer, :text => 3)
    @macview.append_column(col)

    # scrollovatelne treeview
    macscr = Gtk::ScrolledWindow.new
    macscr.set_policy(Gtk::POLICY_AUTOMATIC, Gtk::POLICY_AUTOMATIC)
    macscr.add(@macview)

    @stats = []
    6.times {
      e = Gtk::Entry.new
      e.editable = false
      e.width_chars = 4
      @stats << e
    }

    hbox = Gtk::HBox.new(false)

    @reset = Gtk::Button.new('Reset')
    @reset.signal_connect('clicked') {
      reset
      start
    }

    @stop = Gtk::Button.new('Stop')
    @stop.signal_connect('clicked') {
      if $ports
        stop
        $ports = nil
      end
    }

    hbox.pack_start(@reset, false, false)
    hbox.pack_start(@stop, false, false)

    hbox.pack_start(Gtk::Label.new('L2:'), false, false)
    hbox.pack_start(@stats[0])
    hbox.pack_start(Gtk::Label.new('ICMP:'), false, false)
    hbox.pack_start(@stats[1])
    hbox.pack_start(Gtk::Label.new('IP:'), false, false)
    hbox.pack_start(@stats[2])
    hbox.pack_start(Gtk::Label.new('UDP:'), false, false)
    hbox.pack_start(@stats[3])
    hbox.pack_start(Gtk::Label.new('TCP:'), false, false)
    hbox.pack_start(@stats[4])
    hbox.pack_start(Gtk::Label.new('Ign:'), false, false)
    hbox.pack_start(@stats[5])


    # pridanie treeview
    @packetview = Gtk::TreeView.new(@packetstore)
    renderer = Gtk::CellRendererText.new

    # Add column using our renderer
    col = Gtk::TreeViewColumn.new("DST", renderer, :text => 1)
    col.add_attribute(renderer, "background", 8)
    @packetview.append_column(col)
    col = Gtk::TreeViewColumn.new("SRC", renderer, :text => 2)
    col.add_attribute(renderer, "background", 8)
    @packetview.append_column(col)
    col = Gtk::TreeViewColumn.new("type", renderer, :text => 3)
    col.add_attribute(renderer, "background", 8)
    @packetview.append_column(col)
    col = Gtk::TreeViewColumn.new("DST", renderer, :text => 4)
    col.add_attribute(renderer, "background", 8)
    @packetview.append_column(col)
    col = Gtk::TreeViewColumn.new("Port", renderer, :text => 5)
    col.add_attribute(renderer, "background", 8)
    @packetview.append_column(col)
    col = Gtk::TreeViewColumn.new("SRC", renderer, :text => 6)
    col.add_attribute(renderer, "background", 8)
    @packetview.append_column(col)
    col = Gtk::TreeViewColumn.new("PORT", renderer, :text => 7)
    col.add_attribute(renderer, "background", 8)
    @packetview.append_column(col)

    packetscr = Gtk::ScrolledWindow.new
    packetscr.set_policy(Gtk::POLICY_AUTOMATIC, Gtk::POLICY_AUTOMATIC)
    packetscr.add(@packetview)

    @bufview = Gtk::TextView.new
    @bufview.editable = false
    bufscr = Gtk::ScrolledWindow.new
    bufscr.set_policy(Gtk::POLICY_AUTOMATIC, Gtk::POLICY_AUTOMATIC)
    bufscr.add(@bufview)

    mono = Gtk::TextTag.new('mono')
    mono.font = "Monospace 10"
    @bufview.buffer.tag_table.add(mono)
    @packetview.signal_connect('row-activated') { |view, path, column|
      row = view.selection.selected
      @bufview.buffer.text = ''
      @bufview.buffer.insert(@bufview.buffer.start_iter, row[0].dump, 'mono')
    }

    @rulesview = Gtk::TextView.new
    @rulesview.editable = false
    rulesviewscr = Gtk::ScrolledWindow.new
    rulesviewscr.set_policy(Gtk::POLICY_AUTOMATIC, Gtk::POLICY_AUTOMATIC)
    rulesviewscr.add(@rulesview)

    @rules = Gtk::TextView.new
    @rules.editable = true
    rulesscr = Gtk::ScrolledWindow.new
    rulesscr.set_policy(Gtk::POLICY_AUTOMATIC, Gtk::POLICY_AUTOMATIC)
    rulesscr.add(@rules)

    rulesbutton = Gtk::Button.new('Load rules')
    # is comment line?
    @comment = Regexp.new('^\s*#')

    rulesbutton.signal_connect('clicked') {
      if $ports
        $ports.each { |p| p.filter.clear }
        @rulesview.buffer.text = ''
        @rules.buffer.text.each_line { |l|
          if not @comment.match(l)
            # cut the comment
            rule = Rule.new(l.split('#').first)
            $ports.each { |p|
              if p.dev == rule.dev
                @rulesview.buffer.insert(@rulesview.buffer.end_iter,
                                         "%s\n" % rule.input)
                p.filter.add(rule)
                break
              end
            }
          end
        }

        $ports.each { |p|
          [["input", p.filter.policy_in],
           ["output", p.filter.policy_out]].each { |policy|
          if policy[1] == true
            @rulesview.buffer.insert(@rulesview.buffer.end_iter,
                                     "#{p.dev} default #{policy[0]} policy: ALLOW\n")
          elsif policy[1] == false
            @rulesview.buffer.insert(@rulesview.buffer.end_iter,
                                     "#{p.dev} default #{policy[0]} policy: DENY\n")
          end
          }
        }
      end
    }

    vpan = Gtk::VPaned.new
    vpan.add1(rulesviewscr)
    vpan.add2(rulesscr)

    ruleslabel = Gtk::Label.new("DEVICE in/out [policy allow/deny]\n"\
                                "[icmp/tcp/udp/any]\n"\
                                "[from/to IP/MAC]*\n"\
                                "[from/to tcp/udp/port PORT]*")

    rules = Gtk::VBox.new(false)

    @statsview = Gtk::TreeView.new(@portstats)
    renderer = Gtk::CellRendererText.new
    col = Gtk::TreeViewColumn.new("stat", renderer, :text => 0)
    @statsview.append_column(col)
    col = Gtk::TreeViewColumn.new("bytes", renderer, :text => 1)
    @statsview.append_column(col)
    col = Gtk::TreeViewColumn.new("L2", renderer, :text => 2)
    @statsview.append_column(col)
    col = Gtk::TreeViewColumn.new("ICMP", renderer, :text => 3)
    @statsview.append_column(col)
    col = Gtk::TreeViewColumn.new("IP", renderer, :text => 4)
    @statsview.append_column(col)
    col = Gtk::TreeViewColumn.new("TCP", renderer, :text => 5)
    @statsview.append_column(col)
    col = Gtk::TreeViewColumn.new("UDP", renderer, :text => 6)
    @statsview.append_column(col)

    rules.pack_start(vpan)
    rules.pack_start(ruleslabel, false, false)
    rules.pack_start(rulesbutton, false, false)

    vpan = Gtk::VPaned.new
    vpan.add1(macscr)
    vpan.add2(packetscr)
    vbox = Gtk::VBox.new(false)
    vbox.pack_start(hbox, false, false)
    vbox.pack_start(vpan)


    vpanbuf = Gtk::VPaned.new
    vpanbuf.add1(bufscr)
    vpanbuf.add2(@statsview)

    vpan = Gtk::VPaned.new
    vpan.add1(vpanbuf)
    vpan.add2(rules)

    hpan = Gtk::HPaned.new
    hpan.add1(vbox)
    hpan.add2(vpan)

    # initialize network stuff
    reset_net

    @window.add(hpan)
    @window.show_all
  end

  def reset_tables
    # entry, MAC, Port, Timestamp
    @macstore = Gtk::ListStore.new(LookupEntry,
                                String, String, Integer)
    # data, MAC, MAC, type, IP, port, IP, port, color
    @packetstore = Gtk::ListStore.new(Pcap::Packet,
                                      String, String, String,
                                      String, Integer, String, Integer,
                                      String)
    # in, out
    # bytes, l2, icmp, ip, tcp, udp
    @portstats = Gtk::ListStore.new(String, Integer, Integer, Integer,
                                    Integer, Integer, Integer)
  end

  def reset_net
    stop if $ports
    $ports = []
    $table = LookupTable.new(self)
    $ignored = 0
  end

  def reset
    reset_net
    reset_tables
    @macview.model = @macstore
    @packetview.model = @packetstore
    @statsview.model = @portstats
    @rulesview.buffer.text = ''
  end

  def start
    index = 0

    ARGV.each { |dev|
      port = Port.new(dev, index, @portstats)
      $ports << port
      port.start
      index += 2
    }
  end

  def stop
    $ports.each { |p|
      p.stop
      p.close
    }
    $table.stop if $table
  end
end


# rule: DEVICE in/out [policy allow/deny]
#       [icmp/ip/tcp/udp/any]
#       [ from/to  ip/mac        ADDRESS ]
#       [ from/to  tcp/udp/port  PORT    ]
#
# IP:   ip[/mask]
# MAC:  xxxxxxxx or xx:xx:xx:xx:xx:xx
#
# use match method to match against packet
class Rule
  attr_reader :direction, :dev, :input, :rule, :policy

  def initialize(rule)
    @ip = Regexp.new('^\d+(\.\d+){0,3}$')
    rule.downcase!
    @rule = {}
    # this is mask shift! ip/24 is matched with ip1 >> 8 != ip2 >> 8
    @mask = {}
    # if this is a policy rule
    @policy = nil

    a = rule.split
    if a.length > 2 and ['in', 'out'].include? a[1]
      @dev = a.shift
      @direction = a.shift
      # equivalent of input (parsed)
      @input = "#{@dev} #{@direction.upcase}"

      # policy
      if a[0] == 'policy'
        @input << " POLICY "
        case a[1]
        when 'allow' then
          @policy = true
          @input << "ALLOW"
        when 'deny' then
          @policy = false
          @input << "DENY"
        end
      else
        # match specific protocol
        @l3 = a[0]
        case @l3
        when 'any' then
          @input << " ANY"
          a.shift
        when 'icmp' then
          @input << " ICMP"
          a.shift
        when 'ip' then
          @input << " IP"
          a.shift
        when 'tcp' then
          @input << " TCP"
          a.shift
        when 'udp' then
          @input << " UDP"
          a.shift
        end
        parse(a)
      end
    end
  end

  def parse(a)
    case a.shift
    when 'from' then value(a, 'from')
    when 'to' then value(a, 'to')
    end

    parse(a) if a.length > 0
  end

  def value(a, dir)
    @input << " #{dir.upcase}"
    case a.shift
    when 'ip' then @input << " IP"; ip(a, dir)
    when 'mac' then @input << " MAC"; mac(a, dir)
    when 'tcp' then @input << " TCP"; port(a, dir, 'tcp')
    when 'udp' then @input << " UDP"; port(a, dir, 'udp')
    when 'port' then @input << " PORT"; port(a, dir, 'any')
    end
  end

  def ip(a, dir)
    tmp = a.shift.split('/')
    ip = tmp.shift
    mask = tmp.shift
    if mask
      mask = mask.to_i
    else # no mask is 32 therefore shift is 0
      mask = 32
    end

    # bail if it's bullshit
    if not @ip.match(ip)
      @input << ' *error*'
      return
    end
    ip = Pcap::IPAddress.new(ip)
    @mask[dir] = 32 - mask
    @rule["#{dir}_ip"] = ip
    @input << " #{ip.to_num_s}/#{mask}"
  end

  def port(a, dir, type)
    port = a.shift.to_i
    @rule["#{dir}_port"] = [type, port]
    @input << " #{port}"
  end

  def mac(a, dir)
    mac = num2str(a.shift.split(':').join.to_i(16), 6)
    @rule["#{dir}_mac"] = mac
    @input << " #{sprintf("%02X:%02X:%02X:%02X:%02X:%02X", *mac.unpack("C6"))}"
  end

  def match(pkt)
    r = {}

    # protocol
    case @l3
    when 'icmp' then
      return false if not pkt.icmp?
    when 'ip' then
      # Packet.ip? returns true for ICMP
      return false if (not pkt.ip?) or pkt.icmp?
    when 'tcp' then
      return false if not pkt.tcp?
    when 'udp' then
      return false if not pkt.tcp?
    end

    @rule.each_pair { |k, v|
      case k
      when 'from_mac'
        return false if v != pkt.mac_src
      when 'to_mac'
        return false if v != pkt.mac_dst
      when 'from_ip'
        return false if not pkt.ip?
        if v.to_i >> @mask['from'] != pkt.ip_src.to_i >> @mask['from']
          return false
        end
      when 'to_ip'
        return false if (not pkt.ip?) or v != pkt.ip_dst
        if v.to_i >> @mask['to'] != pkt.ip_dst.to_i >> @mask['to']
          return false
        end
      when 'from_port'
        return false if not (pkt.tcp? or pkt.udp?)
        return false if v[1] != pkt.sport
        if v[0] != 'any'
          return false if v[0] == 'tcp' and pkt.udp?
          return false if v[0] == 'udp' and pkt.tcp?
        end
      when 'to_port'
        return false if not (pkt.tcp? or pkt.udp?)
        return false if v[1] != pkt.dport
        if v[0] != 'any'
          return false if v[0] == 'tcp' and pkt.udp?
          return false if v[0] == 'udp' and pkt.tcp?
        end
      end
    }
    return true
  end
end



if ARGV[0] == nil
  #["eth0", "wmaster0", "wlan0", "any"]
  # 'any' nefunguje
  p Pcap::findalldevs
  exit 0
end

# TODO: teraz flooduje uz len ten multicast
#       eth0 in from mac 01005e7ffffa
#       eth0 in to mac 01005e7ffffa
$local_mac = []
list = Mac.addr.list
p list
list.each { |a|
  $local_mac << (a.split(':').map { |n| n.to_i(16) }).pack("C6")
}

#p Rule.new("eth0 in from ip 127.1 from mac 060504030201 to tcp 80").input
#p Rule.new("eth0 in from ip 127.1 from mac ff:ff:ff:ff:ff:ff to tcp 80")
#p Rule.new("eth0 in from ip 127.1/8 from mac ff:ff:ff:ff:ff:ff to tcp 80")
#exit 0

$window =  MainWindow.new
$window.start
Gtk.main

# note: mozno by sa patrilo learn() dat do mutexu
#       a tiez $ignored counter/update_stats()
